#!/bin/bash

#===============================================================================

# WARNING Сборка будет проводиться в директории, откуда запущен данный скрипт,
# поэтому следует его запускать в отдельной, желательно пустой директории.
#
# Запуск: <...>/build_libatomic.sh VAR=VALUE...
#
# Переменные (в квадратных скобках указаны значения по умолчанию):
#   INSTALL_DIR          - путь установки полной сборки
#   INSTALL_DIR_LIB      - пусь установки библиотек [$INSTALL_DIR/lib]
#   INSTALL_DIR_INCLUDE  - путь установки заголовочных файлов
#                          [$INSTALL_DIR/include]
#   LCC                  - используемый компилятор
#   CFLAGS               - флаги компилятора [-O0]
#   SKIP_DYNAMIC         - пропустить ли сборку динамической библиотеки [0]
#   J                    - количество потоков make [1]
#
# Если указаны INSTALL_DIR_LIB, INSTALL_DIR_INCLUDE, установка lib и include
# будет проведена по этим путям. В противном случае требуется указать переменную
# INSTALL_DIR, на основе которой будут вычислены эти пути.
#
# К CFLAGS принудительно добавляем -fvisibility=protected (см. bug #81005).
# Причём добавляем спереди, чтобы можно было перебить опцией
# -fvisibility=default, если вдруг понадобится.

#===============================================================================

libatomic_init()
{
  build_notify "=> Инициализация"

  GCC_VERSION="9"

  SOURCE_DIR=$THIS_SCRIPT_DIR/gcc-$GCC_VERSION
  BUILD_DIR=$(pwd)

  if [ "${BUILD_DIR:0:${#SOURCE_DIR}}" = "$SOURCE_DIR" ]; then
    build_failure "Сборка должна выполняться вне директории исходников."
  fi

  # Если установлена INSTALL_DIR, два других пути высчитываем на её основе.
  if [ -n "${INSTALL_DIR:+x}" ]; then
    [ -z "${INSTALL_DIR_LIB:+x}" ]
      INSTALL_DIR_LIB=$INSTALL_DIR/lib
    [ -z "${INSTALL_DIR_INCLUDE:+x}" ]
      INSTALL_DIR_INCLUDE=$INSTALL_DIR/include
  fi

  # Пути установки include и lib директорий указываются отдельно. Этот костыль
  # нужен из-за того, что нам нужно что-то передать в аргумент --prefix скрипта
  # configure. Реально по этому пути установлено ничего не будет, т.к. сборка
  # урезана до двух обозначенных директорий. Но вдруг... См. комментарий ниже.
  [ -z "${INSTALL_DIR:+x}" ] && INSTALL_DIR=$BUILD_DIR/TMP-INSTALL

  [ -z "${INSTALL_DIR_LIB:+x}" ] &&
    build_failure "Не установлена переменная INSTALL_DIR_LIB."
  [ -z "${INSTALL_DIR_INCLUDE:+x}" ] &&
    build_failure "Не установлена переменная INSTALL_DIR_INCLUDE."

  # Двоеточие не ставим для тех переменных, которые могут иметь пустые значения.
  [ -z "${LCC:+x}" ] && build_failure "Не установлена переменная LCC."
  [ -z "${CFLAGS+x}" ] && CFLAGS="-O0"
  [ -z "${SKIP_DYNAMIC:+x}" ] && SKIP_DYNAMIC=0
  [ -z "${J+x}" ] && J=1

  # К CFLAGS принудительно добавляем -fvisibility=protected (см. bug #81005)
  # Причём добавляем спереди, чтобы можно было перебить опцией -fvisibility=default
  # (если вдруг понадобится)
  # Отключено из-за bug #142651
#  CFLAGS="-fvisibility=protected $CFLAGS"

  # Мешающие предупреждения
  CFLAGS="$CFLAGS -Wno-error=invalid-memory-model"

  # FIXME
  # Быстрая затычка для bug #114731
  CFLAGS="$CFLAGS -Wno-error=pointer-qual"

  # Пути до исходников и места сборки библиотеки libatomic.
  LIBATOMIC_SOURCE_DIR=$SOURCE_DIR/libatomic
  LIBATOMIC_BUILD_DIR=$BUILD_DIR/libatomic

  # Из-за системы контроля версий даты изменения файлов нарушаются, и при
  # запуске make может быть осуществлена попытка пересоздать "устаревшие"
  # компоненты сборки с помощью autotools. Если следующие переменные окружения
  # установлены, для переконфигурации будут использованы именно они.
  # Устанавливаем их в пустые команды, поскольку реально файлы изменены не были.
  export AUTOCONF=: AUTOHEADER=: AUTOMAKE=: ACLOCAL=:
}

#===============================================================================

libatomic_configure()
{
  build_notify "=> Конфигурация"

  # Архитектуру, на которой проводится сборка, определяем с помощью скрипта.
  build_alias=$($SOURCE_DIR/config.guess)

  # Целевую архитектуру определяем по выдаче компилятора при подаче опции
  # -dumpmachine. Перечислены все возможные на данный момент варианты.
  # NOTE Тут могут быть проблемы при native сборке: значение  host_alias,
  # получаемое ниже, может отличаться от build_alias, полученного с помощью
  # скрипта, и тогда скрипт конфигурации ошибочно определит сборку как cross.
  # Каковы последствия такого несоответствия -- предмет для выяснения.
  host_alias=$($LCC -dumpmachine)
  case $host_alias in
    e2k-linux-gnu)
      host_alias=e2k$($LCC -print-mode=ptr)-mcst-linux-gnu;;
    e2k-kpda) # См. комментарий к макросу ECOMP_QNX_KPDA
      host_alias=e2k-mcst-nto-qnx6.5.0;;
    sparc-linux-gnu)
      host_alias=sparc-unknown-linux-gnu;;
    sparc64-linux-gnu)
      host_alias=sparc64-unknown-linux-gnu;;
    *)
      build_failure "Неизвестный target: $host_alias";;
  esac

  cmd "Создание директории сборки libatomic" \
    mkdir -p $LIBATOMIC_BUILD_DIR

  cmd "Переход в директорию сборки libatomic" \
    cd $LIBATOMIC_BUILD_DIR

  # Параметры конфигурации.
  # --disable-multilib
  #   Отключаем multilib. Не поддерживаются необходимые опции компилятора.
  # --enable-shared, --disable-shared
  #   Подаются в зависимости от того, нужна ли динамеческая версия библиотеки.
  local shared_flag
  [ $SKIP_DYNAMIC -eq 0 ] && shared_flag="--enable-shared"
  [ $SKIP_DYNAMIC -eq 1 ] && shared_flag="--disable-shared"

  cmd "Конфигурация libatomic" \
    $LIBATOMIC_SOURCE_DIR/configure \
      --prefix=$INSTALL_DIR \
      --libdir=$INSTALL_DIR_LIB \
      --build=$build_alias \
      --host=$host_alias \
      --target=$host_alias \
      --disable-multilib \
       $shared_flag \
      CC="$LCC" \
      CFLAGS="$CFLAGS" \
      AR="$($LCC -print-prog-name=ar)" \
      NM="$($LCC -print-prog-name=nm)" \
      STRIP="$($LCC -print-prog-name=strip)" \
      RANLIB="$($LCC -print-prog-name=ranlib)" \
      OBJDUMP="$($LCC -print-prog-name=objdump)"

  cmd "Переход обратно в директорию сборки" \
    cd $BUILD_DIR
}

#===============================================================================

libatomic_build()
{
  cmd "Сборка libatomic" \
    make -C $LIBATOMIC_BUILD_DIR -j$J
}

#===============================================================================

libatomic_install()
{
  # --no-finish передаётся в libtool для того, чтобы после установки библиотек
  # не вызывалась утилита ldconfig.
  cmd "Установка libatomic" \
    make -C $LIBATOMIC_BUILD_DIR -j$J install LIBTOOLFLAGS=--no-finish 
}

#===============================================================================

# Выдавать ошибки об использовании неинициализированных переменных. Для отладки.
set -u

# Определяем путь до этого скрипта. На основе него будут вычислены другие пути.
THIS_SCRIPT_DIR=$(cd "$(dirname "$0")" && pwd)

# Подключаем скрипт, содержащий вспомогательные функции.
source "$THIS_SCRIPT_DIR/../scripts/build.sub"

# Перенаправляем последующий вывод в файл.
build_log $(pwd)/_log

# Сбрасываем значения переменных, которые мы ожидаем получать в качестве
# параметров скрипта. Нужно на случай, если они уже установлены в окружении.
# См. bug #42937.
unset INSTALL_DIR INSTALL_DIR_INCLUDE INSTALL_DIR_LIB LCC CFLAGS SKIP_DYNAMIC J

# Вывод в лог общей информации о сборке и разбор аргументов командной строки.
build_init "libatomic" "$@"

libatomic_init
libatomic_configure
libatomic_build
libatomic_install

build_finish
