#!/bin/bash

#===============================================================================

# WARNING Сборка будет проводиться в директории, откуда запущен данный скрипт,
# поэтому следует его запускать в отдельной, желательно пустой директории.
#
# Запуск: <...>/build_libgcc.sh VAR=VALUE...
#
# Переменные (в квадратных скобках указаны значения по умолчанию):
#   INSTALL_DIR          - путь установки полной сборки
#   INSTALL_DIR_LIB      - пусь установки библиотек [$INSTALL_DIR/lib]
#   INSTALL_DIR_INCLUDE  - путь установки заголовочных файлов
#                          [$INSTALL_DIR/include]
#   LCC                  - используемый компилятор
#   CFLAGS               - флаги компилятора ["-O0"]
#   J                    - количество потоков make [1]
#   STAGE                - номер стадии сборки (см. bug #89506)
#                          1 - установка unwind.h
#                          2 - сборка libgcc.a и libgcc_eh.a
#                          3 - сборка libgcc_s.so
#
# Если указаны INSTALL_DIR_LIB, INSTALL_DIR_INCLUDE, установка lib и include
# будет проведена по этим путям. В противном случае требуется указать переменную
# INSTALL_DIR, на основе которой будут вычислены эти пути.

#===============================================================================

libgcc_init()
{
  GCC_VERSION="9"

  SOURCE_DIR=$THIS_SCRIPT_DIR/gcc-$GCC_VERSION
  BUILD_DIR=$(pwd)

  if [ "${BUILD_DIR:0:${#SOURCE_DIR}}" == "$SOURCE_DIR" ]; then
    build_failure "Сборка должна выполняться вне директории исходников."
  fi

  # Если установлена INSTALL_DIR, два других пути высчитываем на её основе.
  if [ -n "${INSTALL_DIR:+x}" ]; then
    [ -z "${INSTALL_DIR_LIB:+x}" ]
      INSTALL_DIR_LIB=$INSTALL_DIR/lib
    [ -z "${INSTALL_DIR_INCLUDE:+x}" ]
      INSTALL_DIR_INCLUDE=$INSTALL_DIR/include
  fi

  # Пути установки include и lib директорий указываются отдельно. Этот костыль
  # нужен из-за того, что нам нужно что-то передать в аргумент --prefix скрипта
  # configure. Реально по этому пути установлено ничего не будет, т.к. сборка
  # урезана до двух обозначенных директорий. Но вдруг... См. комментарий ниже.
  [ -z "${INSTALL_DIR:+x}" ] && INSTALL_DIR=$BUILD_DIR/TMP-INSTALL

  [ -z "${INSTALL_DIR_LIB:+x}" ] &&
    build_failure "Не установлена переменная INSTALL_DIR_LIB."
  [ -z "${INSTALL_DIR_INCLUDE:+x}" ] &&
    build_failure "Не установлена переменная INSTALL_DIR_INCLUDE."

  # Двоеточие не ставим для тех переменных, которые могут иметь пустые значения.
  [ -z "${LCC:+x}" ] && build_failure "Не установлена переменная LCC."
  [ -z "${CFLAGS+x}" ] && CFLAGS="-O0"
  [ -z "${J+x}" ] && J=1
  [ -z "${STAGE:+x}" ] && build_failure "Не установлена переменная STAGE."

  # Файл содержит всякие штуки, которые должен предоставлять компилятор, но не
  # хочется с этим заморачиваться.
  LCC="$LCC -include $SOURCE_DIR/builtin.h"

  # По этому флагу определяются необходимые макросы.
  CFLAGS="-fbuilding-libgcc -Wno-builtin-declaration-mismatch $CFLAGS"

  GCC_BUILD_DIR=$BUILD_DIR/gcc

  # Пути до исходников и места сборки библиотеки libgcc.
  LIBGCC_SOURCE_DIR=$SOURCE_DIR/libgcc
  LIBGCC_BUILD_DIR=$BUILD_DIR/libgcc

  # Из-за системы контроля версий даты изменения файлов нарушаются, и при
  # запуске make может быть осуществлена попытка пересоздать "устаревшие"
  # компоненты сборки с помощью autotools. Если следующие переменные окружения
  # установлены, для переконфигурации будут использованы именно они.
  # Устанавливаем их в пустые команды, поскольку реально файлы изменены не были.
  export AUTOCONF=: AUTOHEADER=: AUTOMAKE=: ACLOCAL=:
}

#===============================================================================

libgcc_configure()
{
  # Архитектуру, на которой проводится сборка, определяем с помощью скрипта.
  build_alias=$($SOURCE_DIR/config.guess)

  # Целевую архитектуру определяем по выдаче компилятора при подаче опции
  # -dumpmachine. Перечислены все возможные на данный момент варианты.
  # NOTE Тут могут быть проблемы при native сборке: значение  host_alias,
  # получаемое ниже, может отличаться от build_alias, полученного с помощью
  # скрипта, и тогда скрипт конфигурации ошибочно определит сборку как cross.
  # Каковы последствия такого несоответствия -- предмет для выяснения.
  host_alias=$($LCC -dumpmachine)
  case $host_alias in
    e2k-linux-gnu)
      host_alias=e2k-mcst-linux-gnu;;
    e2k-kpda) # См. комментарий к макросу ECOMP_QNX_KPDA
	# Для e2k-kpda при конфигурации libgcc пока используем тот же `--host',
	# что и для e2k-linux (см. Bug #97006, Comment #4).
      host_alias=e2k-mcst-linux-gnu;;
    sparc-linux-gnu)
      host_alias=sparc-unknown-linux-gnu;;
    sparc64-linux-gnu)
      host_alias=sparc64-unknown-linux-gnu;;
    *)
      build_failure "Неизвестный target: $host_alias";;
  esac

  # Имитация конфигурации gcc.
  cmd "Создание директории сборки gcc" \
    mkdir -p $GCC_BUILD_DIR

  cmd "Копирование конфигурационных файлов gcc" \
    cp $SOURCE_DIR/gcc_objdir/* $GCC_BUILD_DIR

  # Собственно, конфигурация.
  cmd "Создание директории сборки libgcc" \
    mkdir -p $LIBGCC_BUILD_DIR

  cmd "Переход в директорию сборки libgcc" \
    cd $LIBGCC_BUILD_DIR

  # Параметры конфигурации:
  # --with-cross-host
  #   Параметр не указан намеренно, поскольку при его указании установка
  #   библиотек происходит по неудобным нам путям. Больше этот параметр вроде бы
  #   ни на что не влияет.
  # --disable-sjlj-exceptions
  #   Явно указываем скрипту, что мы не хотим SJLJ-модель исключений. Сам он её
  #   определить не в состоянии, поскольку у нас EDG реализация исключений.
  # --disable-multilib
  #   Отключаем multilib. Не поддерживаются необходимые опции компилятора.
  # --disable-decimal-float
  #   Отключение функций, работающих с типом данных _Decimal. Не поддерживается.
  cmd "Конфигурация libgcc" \
    $LIBGCC_SOURCE_DIR/configure \
      --prefix=$INSTALL_DIR \
      --build=$build_alias \
      --host=$host_alias \
      --target=$host_alias \
      --disable-sjlj-exceptions \
      --disable-multilib \
      --disable-decimal-float \
      CC="$LCC" \
      CFLAGS="$CFLAGS" \
      AR="$($LCC -print-prog-name=ar)" \
      AS="$($LCC -print-prog-name=as)" \
      NM="$($LCC -print-prog-name=nm)" \
      STRIP="$($LCC -print-prog-name=strip)" \
      RANLIB="$($LCC -print-prog-name=ranlib)" \
      CXXFILT="$($LCC -print-prog-name=c++filt)" \
      OBJDUMP="$($LCC -print-prog-name=objdump)"

  cmd "Переход обратно в директорию сборки" \
    cd $BUILD_DIR
}

#===============================================================================

libgcc_phase1()
{
  cmd "Создание директории установки include-файлов" \
    mkdir -p $INSTALL_DIR_INCLUDE

  cmd "Генерация unwind.h" \
    make -C $LIBGCC_BUILD_DIR -j$J gcc_objdir=$GCC_BUILD_DIR unwind.h

  cmd "Установка unwind.h" \
    cp -f $LIBGCC_BUILD_DIR/unwind.h $INSTALL_DIR_INCLUDE
}

#===============================================================================

libgcc_phase2()
{
  cmd "Сборка CRT модулей" \
    make -C $LIBGCC_BUILD_DIR -j$J gcc_objdir=$GCC_BUILD_DIR libgcc-extra-parts

  cmd "Сборка библиотеки libgcc.a" \
    make -C $LIBGCC_BUILD_DIR -j$J gcc_objdir=$GCC_BUILD_DIR libgcc.a

#   cmd "Сборка библиотеки libgcov.a" \
#     make -C $LIBGCC_BUILD_DIR -j$J gcc_objdir=$GCC_BUILD_DIR libgcov.a

  cmd "Сборка библиотеки libgcc_eh.a" \
    make -C $LIBGCC_BUILD_DIR -j$J gcc_objdir=$GCC_BUILD_DIR libgcc_eh.a

  cmd "Создание директории установки" \
    mkdir -p $INSTALL_DIR_LIB

  cmd "Установка общих CRT модулей" \
    cp -f $LIBGCC_BUILD_DIR/crtbegin.o $LIBGCC_BUILD_DIR/crtend.o \
      $LIBGCC_BUILD_DIR/crtbeginS.o $LIBGCC_BUILD_DIR/crtendS.o \
      $LIBGCC_BUILD_DIR/crtbeginT.o $INSTALL_DIR_LIB

  if [ -f $LIBGCC_BUILD_DIR/crtalign.o ]; then
    cmd "Установка crtalign.o" \
      cp -f $LIBGCC_BUILD_DIR/crtalign.o $INSTALL_DIR_LIB
  fi

  if [ -f $LIBGCC_BUILD_DIR/crtunaligned.o ]; then
    cmd "Установка crtunaligned.o" \
      cp -f $LIBGCC_BUILD_DIR/crtunaligned.o $INSTALL_DIR_LIB
  fi

  cmd "Установка libgcc.a" \
    cp -f $LIBGCC_BUILD_DIR/libgcc.a $INSTALL_DIR_LIB

#   cmd "Установка libgcov.a" \
#     cp -f $LIBGCC_BUILD_DIR/libgcov.a $INSTALL_DIR_LIB

  cmd "Установка libgcc_eh.a" \
    cp -f $LIBGCC_BUILD_DIR/libgcc_eh.a $INSTALL_DIR_LIB
}

#===============================================================================

libgcc_phase3()
{
  cmd "Сборка libgcc_s.so" \
    make -C $LIBGCC_BUILD_DIR -j$J gcc_objdir=$GCC_BUILD_DIR libgcc_s.so

  cmd "Установка libgcc_s.so" \
    cp -f $LIBGCC_BUILD_DIR/libgcc_s.so.1 $INSTALL_DIR_LIB

  cmd "Создание ссылки libgcc_s.so -> libgcc_s.so.1" \
    ln -sf libgcc_s.so.1 $INSTALL_DIR_LIB/libgcc_s.so
}

#===============================================================================

# Выдавать ошибки об использовании неинициализированных переменных. Для отладки.
set -u

# Определяем путь до этого скрипта. На основе него будут вычислены другие пути.
THIS_SCRIPT_DIR=$(cd "$(dirname "$0")" && pwd)

# Подключаем скрипт, содержащий вспомогательные функции.
source "$THIS_SCRIPT_DIR/../scripts/build.sub"

# Перенаправляем последующий вывод в файл.
build_log $(pwd)/_log

# Сбрасываем значения переменных, которые мы ожидаем получать в качестве
# параметров скрипта. Нужно на случай, если они уже установлены в окружении.
# См. bug #42937.
unset INSTALL_DIR INSTALL_DIR_INCLUDE INSTALL_DIR_LIB LCC CFLAGS J
unset STAGE

# Вывод в лог общей информации о сборке и разбор аргументов командной строки.
build_init "libgcc" "$@"

libgcc_init
[ $STAGE = 1 ] && libgcc_configure
[ $STAGE = 1 ] && libgcc_phase1
[ $STAGE = 2 ] && libgcc_phase2
[ $STAGE = 3 ] && libgcc_phase3

build_finish
