#!/bin/bash

#===============================================================================

# WARNING Сборка будет проводиться в директории, откуда запущен данный скрипт,
# поэтому следует его запускать в отдельной, желательно пустой директории.
#
# Запуск: <...>/build_libstdc++.sh VAR=VALUE...
#
# Переменные (в квадратных скобках указаны значения по умолчанию):
#   INSTALL_DIR          - путь установки полной сборки
#   INSTALL_DIR_LIB      - пусь установки библиотек [$INSTALL_DIR/lib]
#   INSTALL_DIR_INCLUDE  - путь установки заголовочных файлов
#                          [$INSTALL_DIR/include/c++/g++-v3]
#   LCC                  - используемый компилятор
#   CFLAGS               - флаги компилятора [-O0]
#   SKIP_DYNAMIC         - пропустить ли сборку динамической библиотеки [0]
#   J                    - количество потоков make [1]
#
# Если указаны INSTALL_DIR_LIB, INSTALL_DIR_INCLUDE, установка lib и include
# будет проведена по этим путям. В противном случае требуется указать переменную
# INSTALL_DIR, на основе которой будут вычислены эти пути.
#
# К CFLAGS принудительно добавляем -fvisibility=protected (см. bug #81005)
# Причём добавляем спереди, чтобы можно было перебить опцией -fvisibility=default
# (если вдруг понадобится)

#===============================================================================

libstdcxx_init()
{
  build_notify "=> Инициализация"

  GCC_VERSION="9"

  SOURCE_DIR=$THIS_SCRIPT_DIR/gcc-$GCC_VERSION
  BUILD_DIR=$(pwd)

  if [ "${BUILD_DIR:0:${#SOURCE_DIR}}" = "$SOURCE_DIR" ]; then
    build_failure "Сборка должна выполняться вне директории исходников."
  fi

  # Если установлена INSTALL_DIR, два других пути высчитываем на её основе.
  if [ -n "${INSTALL_DIR:+x}" ]; then
    [ -z "${INSTALL_DIR_LIB:+x}" ]
      INSTALL_DIR_LIB=$INSTALL_DIR/lib
    [ -z "${INSTALL_DIR_INCLUDE:+x}" ]
      INSTALL_DIR_INCLUDE=$INSTALL_DIR/include/c++/g++-v3
  fi

  # Пути установки include и lib директорий указываются отдельно. Этот костыль
  # нужен из-за того, что нам нужно что-то передать в аргумент --prefix скрипта
  # configure. Реально по этому пути установлено ничего не будет, т.к. сборка
  # урезана до двух обозначенных директорий. Но вдруг... См. комментарий ниже.
  [ -z "${INSTALL_DIR:+x}" ] && INSTALL_DIR=$BUILD_DIR/TMP-INSTALL

  [ -z "${INSTALL_DIR_LIB:+x}" ] &&
    build_failure "Не установлена переменная INSTALL_DIR_LIB."
  [ -z "${INSTALL_DIR_INCLUDE:+x}" ] &&
    build_failure "Не установлена переменная INSTALL_DIR_INCLUDE."

  # Двоеточие не ставим для тех переменных, которые могут иметь пустые значения.
  [ -z "${LCC:+x}" ] && build_failure "Не установлена переменная LCC."
  [ -z "${CFLAGS+x}" ] && CFLAGS="-O0"
  [ -z "${SKIP_DYNAMIC:+x}" ] && SKIP_DYNAMIC=0
  [ -z "${J+x}" ] && J=1

  # FIXME Затычка для bug #122813.32 пункт 5 (bug #123770)
#  [ "$GCC_VERSION" = "9" ] && CFLAGS="$CFLAGS -fno-gnu-unique"

  # К CFLAGS принудительно добавляем -fvisibility=protected (см. bug #81005)
  # Причём добавляем спереди, чтобы можно было перебить опцией -fvisibility=default
  # (если вдруг понадобится)
  # Отключено из-за bug #142651
#  CFLAGS="-fvisibility=protected $CFLAGS"

  # Пути до исходников и места сборки библиотеки libstdc++.
  LIBSTDCXX_SOURCE_DIR=$SOURCE_DIR/libstdc++-v3
  LIBSTDCXX_BUILD_DIR=$BUILD_DIR/libstdc++-v3

  # Пути до исходников и места сборки библиотеки libgcc. Нужны для читов.
  LIBGCC_SOURCE_DIR=$SOURCE_DIR/libgcc
  LIBGCC_BUILD_DIR=$BUILD_DIR/libgcc

  # Из-за системы контроля версий даты изменения файлов нарушаются, и при
  # запуске make может быть осуществлена попытка пересоздать "устаревшие"
  # компоненты сборки с помощью autotools. Если следующие переменные окружения
  # установлены, для переконфигурации будут использованы именно они.
  # Устанавливаем их в пустые команды, поскольку реально файлы изменены не были.
  export AUTOCONF=: AUTOHEADER=: AUTOMAKE=: ACLOCAL=:
}

#===============================================================================

libstdcxx_configure()
{
  build_notify "=> Конфигурация"

  # Архитектуру, на которой проводится сборка, определяем с помощью скрипта.
  build_alias=$($SOURCE_DIR/config.guess)

  # Целевую архитектуру определяем по выдаче компилятора при подаче опции
  # -dumpmachine. Перечислены все возможные на данный момент варианты.
  # NOTE Тут могут быть проблемы при native сборке: значение  host_alias,
  # получаемое ниже, может отличаться от build_alias, полученного с помощью
  # скрипта, и тогда скрипт конфигурации ошибочно определит сборку как cross.
  # Каковы последствия такого несоответствия -- предмет для выяснения.
  host_alias=$($LCC -dumpmachine)
  case $host_alias in
    e2k-linux-gnu)
      host_alias=e2k-mcst-linux-gnu;;
    e2k-kpda) # См. комментарий к макросу ECOMP_QNX_KPDA
      host_alias=e2k-mcst-nto-qnx6.5.0;;
    sparc-linux-gnu)
      host_alias=sparc-unknown-linux-gnu;;
    sparc64-linux-gnu)
      host_alias=sparc64-unknown-linux-gnu;;
    *)
      build_failure "Неизвестный target: $host_alias";;
  esac

  # Имитация конфигурации libgcc.
  cmd "Создание директории сборки libgcc" \
    mkdir -p $LIBGCC_BUILD_DIR

  cmd "Создание ссылки gthr-default.h -> gtrh-posix.h" \
    ln -sf $LIBGCC_SOURCE_DIR/gthr-posix.h $LIBGCC_BUILD_DIR/gthr-default.h

  cmd "Создание директории сборки libstdc++" \
    mkdir -p $LIBSTDCXX_BUILD_DIR

  cmd "Переход в директорию сборки libstdc++" \
    cd $LIBSTDCXX_BUILD_DIR

  # Параметры конфигурации.
  # --with-cross-host
  #   Параметр не указан намеренно, поскольку при его указании установка
  #   библиотек происходит по неудобным нам путям. Больше этот параметр вроде бы
  #   ни на что не влияет.
  # --disable-multilib
  #   Отключаем multilib. Не поддерживаются необходимые опции компилятора.
  # --disable-nls
  #   Отключаем многоязыковую поддержку. Она включает в себя перевод пары строк
  #   на французский и немецкий языки. Нам оно ни к чему.
  # --disable-libstdcxx-pch
  #   Отключаем попытки сборки с использованием PCH. У нас не поддерживается.
  # --enable-shared, --disable-shared
  #   Подаются в зависимости от того, нужна ли динамеческая версия библиотеки.
  # CXX и CXXFLAGS устанавливаем так, как это делает GCC.

  # writev () turns out to be broken in "linux-314.13.{xy >= 78}",
  # therefore avoid its usage in PM libstdc++ for now (Bug #107300).
  local writev_flag
  if [ "$($LCC -print-mode=ptr)" = "128" ]; then
    writev_flag="glibcxx_cv_WRITEV=no"
  else
    writev_flag=""
  fi

  local shared_flag
  if [ $SKIP_DYNAMIC -eq 0 ]; then
    shared_flag="--enable-shared"
  else
    shared_flag="--disable-shared"
  fi

  cmd "Конфигурация libstdc++" \
    $LIBSTDCXX_SOURCE_DIR/configure \
      --prefix=$INSTALL_DIR \
      --libdir=$INSTALL_DIR_LIB \
      --with-gxx-include-dir=$INSTALL_DIR_INCLUDE \
      --build=$build_alias \
      --host=$host_alias \
      --target=$host_alias \
      --disable-multilib \
      --disable-nls \
      --disable-libstdcxx-pch \
      $shared_flag \
      $writev_flag \
      CC="$LCC" \
      CFLAGS="$CFLAGS" \
      CXX="$LCC -shared-libgcc -nostdinc++" \
      CXXFLAGS="$CFLAGS -D_GNU_SOURCE" \
      AR="$($LCC -print-prog-name=ar)" \
      AS="$($LCC -print-prog-name=as)" \
      NM="$($LCC -print-prog-name=nm)" \
      STRIP="$($LCC -print-prog-name=strip)" \
      RANLIB="$($LCC -print-prog-name=ranlib)" \
      CXXFILT="$($LCC -print-prog-name=c++filt)" \
      OBJDUMP="$($LCC -print-prog-name=objdump)"

  cmd "Переход обратно в директорию сборки" \
    cd $BUILD_DIR
}

#===============================================================================

libstdcxx_build()
{
  cmd "Сборка libstdc++" \
    make -C $LIBSTDCXX_BUILD_DIR -j$J
}

#===============================================================================

libstdcxx_install()
{
  local pointer_mode=$($LCC -print-mode=ptr)
  local ml_subdir=/ptr$pointer_mode

  # Имитируем включение multilib'a, чтобы установка архитектурно-зависимых
  # заголовочных файлов прошла в $gxx_include_dir/$host_alias$ml_subdir/bits.
  # Подробнее в config-ml.in и libstdc++-v3/include/Makefile.am
  # NOTE В квадратных скобках пробел и символ горизонтальной табуляции.
  cmd "Имитация включения multilib'а" \
    sed -i "s:^MULTISUBDIR[ 	]*=.*$:MULTISUBDIR = ${ml_subdir}:" \
      $LIBSTDCXX_BUILD_DIR/include/Makefile

  # --no-finish передаётся в libtool для того, чтобы после установки библиотек
  # не вызывалась утилита ldconfig.
  cmd "Установка libstdc++" \
    make -C $LIBSTDCXX_BUILD_DIR -j$J install LIBTOOLFLAGS=--no-finish

  cmd "Создание ссылки на архитектурно зависимый каталог" \
    ln -sf $host_alias$ml_subdir $INSTALL_DIR_INCLUDE/tdep.ptr$pointer_mode

  # Для внутренней конфигурации под libmcst временно создаем фейковые
  # <{io,f}stream.h> ради дремучих защищенных e2kt'шных тестов, всю жизнь
  # собиравшихся с libstlport.
  if [ "$host_alias" = "e2k-mcst-linux-gnu" ]; then
    cat <<EOF > $INSTALL_DIR_INCLUDE/iostream.h
#if defined __ptr128__
using namespace std;
#include <iostream>
#endif
EOF

    cat <<EOF > $INSTALL_DIR_INCLUDE/fstream.h
#if defined __ptr128__
#include <fstream>
#endif
EOF
  fi
}

#===============================================================================

# Выдавать ошибки об использовании неинициализированных переменных. Для отладки.
set -u

# Определяем путь до этого скрипта. На основе него будут вычислены другие пути.
THIS_SCRIPT_DIR=$(cd "$(dirname "$0")" && pwd)

# Подключаем скрипт, содержащий вспомогательные функции.
source "$THIS_SCRIPT_DIR/../scripts/build.sub"

# Перенаправляем последующий вывод в файл.
build_log $(pwd)/_log

# Сбрасываем значения переменных, которые мы ожидаем получать в качестве
# параметров скрипта. Нужно на случай, если они уже установлены в окружении.
# См. bug #42937.
unset INSTALL_DIR INSTALL_DIR_INCLUDE INSTALL_DIR_LIB LCC CFLAGS SKIP_DYNAMIC J

# Вывод в лог общей информации о сборке и разбор аргументов командной строки.
build_init "libstdc++" "$@"

libstdcxx_init
libstdcxx_configure
libstdcxx_build
libstdcxx_install

build_finish
