#!/bin/bash

#===============================================================================

# WARNING Сборка будет проводиться в директории, откуда запущен данный скрипт,
# поэтому следует его запускать в отдельной, желательно пустой директории.
#
# Запуск: <...>/build.sh VAR=VALUE...
#
# Переменные (в квадратных скобках указаны значения по умолчанию):
#   INSTALL_DIR_LIB      - пусь установки библиотек
#   INSTALL_DIR_INCLUDE  - путь установки заголовочных файлов
#   LCC                  - используемый компилятор языка C
#   LXX                  - используемый компилятор языка C++
#   CFLAGS               - флаги компилятора [-O0]
#   J                    - количество потоков make [1]
#   DEBUG                - включить сборку в отладочном режиме.
#                          Допустимые значения: ON, OFF [OFF]
#
#===============================================================================

compiler_rt_init()
{
    build_notify "=> Инициализация"

    [ -z "${INSTALL_DIR_LIB:+x}" ] &&
        build_failure "Не установлена переменная INSTALL_DIR_LIB."

    [ -z "${INSTALL_DIR_INCLUDE:+x}" ] &&
        build_failure "Не установлена переменная INSTALL_DIR_INCLUDE."

    # Двоеточие не ставим для тех переменных, которые могут иметь пустые значения.
    [ -z "${LCC:+x}"          ] && build_failure "Не установлена переменная LCC."
    [ -z "${LXX:+x}"          ] && build_failure "Не установлена переменная LXX."
    [ -z "${CFLAGS+x}"        ] && CFLAGS="-O0"
    [ -z "${J+x}"             ] && J=1
    [ -z "${DEBUG+x}"         ] && DEBUG="OFF"

    if [ "$DEBUG" != "ON" -a "$DEBUG" != "OFF" ]
    then
        build_failure "Недопустимые значения для опции DEBUG. Возможные значения: ON, OFF"
    fi

    # Библиотека существует только в статическом варианте, поэтому требуются
    # дополнительные опции, чтобы её можно нормально прилинковать
    # к пользовательским динамическим библиотекам
    #
    # Аналог CFLAGS_STATIC_ONLY в сборках, собираемых через Makefile
    CFLAGS="$CFLAGS -fPIC -fvisibility=hidden"

    # Мешающие предупреждения
    CFLAGS="$CFLAGS -Wno-error=unused-but-set-variable"

    # Есть опции, которые по смыслу должны быть частью компилятора
    # (например, определяющие битность целевой архитектуры), и все
    # исходные файлы должны компилироваться обязательно с этими опциями.
    # Поэтому в качестве LCC/LXX подается компилятор + такие обязательные
    # опции, но cmake ругается, если в качестве компилятора ему подается
    # еще что-то. По этой причине формируем скрипт, в котором происходит
    # запуск компилятора с обязательными опциями, и этот скрипт кормим
    # cmake'у в качестве компилятора

    # Для старой версии cmake 2.4 передаём все флаги через скрипт.
    CMAKE_FLAGS=
    cmake --version | head -1 | grep '2.4' > /dev/null && CMAKE_FLAGS=$CFLAGS

    LCC_COMP="$BUILD_DIR/lcc_comp"
    LXX_COMP="$BUILD_DIR/lxx_comp"
    echo -e '#!/bin/bash\n'$LCC $CMAKE_FLAGS' $@' > $LCC_COMP
    echo -e '#!/bin/bash\n'$LXX $CMAKE_FLAGS' $@' > $LXX_COMP
    chmod +x $LCC_COMP
    chmod +x $LXX_COMP

    # Определяем нужно ли подавать опцию -Werror. Во время сборки 32-битной
    # библиотеки есть множество предупреждений вида "conversion from integer
    # to smaller pointer". Поэтому пока не появится опция, отключающая данный
    # варнинг, -Werror подаем только при сборке 64-битной библиотеки
    # FIXME: после исправления bug 101400 добавить -Werror для всех сборок
    IS_WERROR="ON"

    local m32="\-mptr32\|\-m32"

    echo $CFLAGS | grep "$m32" > /dev/null && IS_WERROR="OFF"
    echo $LCC    | grep "$m32" > /dev/null && IS_WERROR="OFF"
    echo $LXX    | grep "$m32" > /dev/null && IS_WERROR="OFF"

    # Из-за системы контроля версий даты изменения файлов нарушаются, и при
    # запуске make может быть осуществлена попытка пересоздать "устаревшие"
    # компоненты сборки с помощью autotools. Если следующие переменные окружения
    # установлены, для переконфигурации будут использованы именно они.
    # Устанавливаем их в пустые команды, поскольку реально файлы изменены не были.
    export AUTOCONF=: AUTOHEADER=: AUTOMAKE=: ACLOCAL=:
}

#===============================================================================

compiler_rt_configure()
{
    build_notify "=> Конфигурация"

    cmd "Конфигурация compiler_rt" \
        cmake \
        -DMCST:BOOL=TRUE \
        -DCOMPILER_RT_ENABLE_WERROR:BOOL=$IS_WERROR \
        -DMCST_SANITIZE_LIBRARY_INSTALL_DIR:PATH=$INSTALL_DIR_LIB \
        -DMCST_SANITIZE_HEADER_INSTALL_DIR:PATH=$INSTALL_DIR_INCLUDE \
        -DCMAKE_C_COMPILER=$LCC_COMP \
        -DCMAKE_CXX_COMPILER=$LXX_COMP \
        -DCMAKE_C_FLAGS="$CFLAGS" \
        -DCMAKE_CXX_FLAGS="$CFLAGS" \
        -DCOMPILER_RT_DEBUG:BOOL=$DEBUG \
        $SOURCE_DIR
}

#===============================================================================

compiler_rt_build()
{
    build_notify "=> Сборка"

    cmd "Сборка compiler_rt" \
        make -j$J VERBOSE=1
}

#===============================================================================

compiler_rt_install()
{
    build_notify "=> Установка"

    cmd "Установка compiler_rt" \
        make install -j$J
}

#===============================================================================

# Выдавать ошибки об использовании неинициализированных переменных. Для отладки.
set -u

# Определяем путь до этого скрипта. На основе него будут вычислены другие пути.
THIS_SCRIPT_DIR=$(cd "$(dirname "$0")" && pwd)

# Определяем основные пути
SOURCE_DIR=$THIS_SCRIPT_DIR
BUILD_DIR=$(pwd)

# Подключаем скрипт, содержащий вспомогательные функции.
source "$THIS_SCRIPT_DIR/../scripts/build.sub"

# Перенаправляем последующий вывод в файл.
build_log $(pwd)/_log

# Сбрасываем значения переменных, которые мы ожидаем получать в качестве
# параметров скрипта. Нужно на случай, если они уже установлены в окружении.
# См. bug #42937.
unset INSTALL_DIR_LIB LCC LXX CFLAGS J

# Вывод в лог общей информации о сборке и разбор аргументов командной строки.
build_init "compiler_rt" "$@"

compiler_rt_init
compiler_rt_configure
compiler_rt_build
compiler_rt_install

build_finish
