include(CMakeParseArguments)
include(CompilerRTUtils)

macro(set_target_output_directories target output_dir)
  # For RUNTIME_OUTPUT_DIRECTORY variable, Multi-configuration generators
  # append a per-configuration subdirectory to the specified directory.
  # To avoid the appended folder, the configuration specific variable must be
  # set 'RUNTIME_OUTPUT_DIRECTORY_${CONF}':
  # RUNTIME_OUTPUT_DIRECTORY_DEBUG, RUNTIME_OUTPUT_DIRECTORY_RELEASE, ...
  if(CMAKE_CONFIGURATION_TYPES)
    foreach(build_mode ${CMAKE_CONFIGURATION_TYPES})
      string(TOUPPER "${build_mode}" CONFIG_SUFFIX)
      set_target_properties("${target}" PROPERTIES
          "ARCHIVE_OUTPUT_DIRECTORY_${CONFIG_SUFFIX}" ${output_dir}
          "LIBRARY_OUTPUT_DIRECTORY_${CONFIG_SUFFIX}" ${output_dir}
          "RUNTIME_OUTPUT_DIRECTORY_${CONFIG_SUFFIX}" ${output_dir})
    endforeach(build_mode)
  else(CMAKE_CONFIGURATION_TYPES)
    set_target_properties("${target}" PROPERTIES
        ARCHIVE_OUTPUT_DIRECTORY ${output_dir}
        LIBRARY_OUTPUT_DIRECTORY ${output_dir}
        RUNTIME_OUTPUT_DIRECTORY ${output_dir})
  endif(CMAKE_CONFIGURATION_TYPES)
endmacro(set_target_output_directories)

# MCST: add prefix to all emelemts of list and write to out_val string
macro(add_prefix out_val prefix in_list)
  set(${out_val} "")
  foreach(def ${in_list})
    set(${out_val} "${_defines} ${prefix}${def}")
  endforeach(def)
endmacro(add_prefix)

# Tries to add an "object library" target for a given list of OSs and/or
# architectures with name "<name>.<arch>" for non-Darwin platforms if
# architecture can be targeted, and "<name>.<os>" for Darwin platforms.
# add_compiler_rt_object_libraries(<name>
#                                  OS <os names>
#                                  ARCHS <architectures>
#                                  SOURCES <source files>
#                                  CFLAGS <compile flags>
#                                  DEFS <compile definitions>)
macro(add_compiler_rt_object_libraries name)
  cmake_parse_arguments(LIB "" "" "ARCHS;SOURCES;CFLAGS;DEFS" ${ARGN})
  set(libnames)
  
  foreach(arch ${LIB_ARCHS})
    set(libname "${name}.${arch}")
    set(libnames ${libnames} ${libname})
    set(extra_cflags_${libname} ${TARGET_${arch}_CFLAGS})
    if(NOT CAN_TARGET_${arch})
      message(FATAL_ERROR "Architecture ${arch} can't be targeted")
    endif(NOT CAN_TARGET_${arch})
  endforeach(arch)

  # MCST: append preprocessor prefix for all defines
  add_prefix(_defines "-D" "${LIB_DEFS}")

  foreach(libname ${libnames})
    # add_library(${libname} OBJECT ${LIB_SOURCES})
    # MCST: older version of cmake doesn't support object library
    add_library(${libname} STATIC ${LIB_SOURCES})
    set_target_compile_flags(${libname}
      ${CMAKE_CXX_FLAGS} ${extra_cflags_${libname}} ${LIB_CFLAGS} ${_defines})
  endforeach(libname)
endmacro(add_compiler_rt_object_libraries)

# Takes a list of object library targets, and a suffix and appends the proper
# TARGET_OBJECTS string to the output variable.
# format_object_libs(<output> <suffix> ...)
macro(format_object_libs output suffix)
  foreach(lib ${ARGN})
    # list(APPEND ${output} $<TARGET_OBJECTS:${lib}.${suffix}>)
    # MCST: older version of cmake doesn't support object library
    list(APPEND ${output} ${lib}.${suffix})
  endforeach(lib)
endmacro(format_object_libs)

# Adds static or shared runtime for a list of architectures and operating
# systems and puts it in the proper directory in the build and install trees.
# add_compiler_rt_runtime(<name>
#                         {STATIC|SHARED}
#                         ARCHS <architectures>
#                         OS <os list>
#                         SOURCES <source files>
#                         CFLAGS <compile flags>
#                         LINKFLAGS <linker flags>
#                         DEFS <compile definitions>
#                         LINK_LIBS <linked libraries> (only for shared library)
#                         OBJECT_LIBS <object libraries to use as sources>
#                         LANG <language for library>
#                         PARENT_TARGET <convenience parent target>)
macro(add_compiler_rt_runtime name type)
  if(NOT ${type} MATCHES "^(STATIC|SHARED)$")
    message(FATAL_ERROR "type argument must be STATIC or SHARED")
  endif(NOT ${type} MATCHES "^(STATIC|SHARED)$")
  cmake_parse_arguments(LIB
    ""
    "PARENT_TARGET"
    "OS;ARCHS;SOURCES;CFLAGS;LINKFLAGS;DEFS;LINK_LIBS;OBJECT_LIBS;LANG"
    ${ARGN})
  set(libnames)
  foreach(arch ${LIB_ARCHS})
    if(NOT CAN_TARGET_${arch})
      message(FATAL_ERROR "Architecture ${arch} can't be targeted")
    endif(NOT CAN_TARGET_${arch})
    if(${type} STREQUAL "STATIC")
      #set(libname "${name}-${arch}")
      # MCST: remove the arch from target name
      set(libname ${name})
      set(output_name_${libname} ${libname}${COMPILER_RT_OS_SUFFIX})
    else(${type} STREQUAL "STATIC")
      set(libname "${name}-dynamic-${arch}")
      set(extra_cflags_${libname} ${TARGET_${arch}_CFLAGS} ${LIB_CFLAGS})
      set(extra_linkflags_${libname} ${TARGET_${arch}_LINKFLAGS} ${LIB_LINKFLAGS})
      set(output_name_${libname} ${name}-${arch}${COMPILER_RT_OS_SUFFIX})
    endif(${type} STREQUAL "STATIC")
    set(sources_${libname} ${LIB_SOURCES})
    format_object_libs(libs_${libname} ${arch} ${LIB_OBJECT_LIBS})
    set(libnames ${libnames} ${libname})
    set(extra_cflags_${libname} ${TARGET_${arch}_CFLAGS} ${LIB_CFLAGS})
  endforeach(arch)

  if(libnames)
    if(LIB_PARENT_TARGET)
      # The parent install target specifies the parent component to scrape up
      # anything not installed by the individual install targets, and to handle
      # installation when running the multi-configuration generators.
      add_custom_target(install-${LIB_PARENT_TARGET}
                        DEPENDS ${LIB_PARENT_TARGET}
                        COMMAND "${CMAKE_COMMAND}"
                                -DCMAKE_INSTALL_COMPONENT=${LIB_PARENT_TARGET}
                                -P "${CMAKE_BINARY_DIR}/cmake_install.cmake")
    endif(LIB_PARENT_TARGET)

    foreach(libname ${libnames})
      # If you are using a multi-configuration generator we don't generate
      # per-library install rules, so we fall back to the parent target COMPONENT
      if(CMAKE_CONFIGURATION_TYPES AND LIB_PARENT_TARGET)
        set(COMPONENT_OPTION COMPONENT ${LIB_PARENT_TARGET})
      else(CMAKE_CONFIGURATION_TYPES AND LIB_PARENT_TARGET)
        set(COMPONENT_OPTION COMPONENT ${libname})
      endif(CMAKE_CONFIGURATION_TYPES AND LIB_PARENT_TARGET)

      add_library(${libname} ${type} ${sources_${libname}})

      if (LIB_LANG)
        set_target_properties(${libname} PROPERTIES LINKER_LANGUAGE ${LIB_LANG})
      endif(LIB_LANG)

      target_link_libraries(${libname} ${libs_${libname}})
      # MCST: append preprocessor prefix for all defines
      add_prefix(_defines "-D" "${LIB_DEFS}")
      set_target_compile_flags(${libname} ${extra_cflags_${libname}} ${_defines})
      set_target_link_flags(${libname} ${extra_linkflags_${libname}})

      set_target_output_directories(${libname} ${COMPILER_RT_LIBRARY_OUTPUT_DIR})
      set_target_properties(${libname} PROPERTIES
          OUTPUT_NAME ${output_name_${libname}})

      if(LIB_LINK_LIBS AND ${type} STREQUAL "SHARED")
        target_link_libraries(${libname} ${LIB_LINK_LIBS})
      endif(LIB_LINK_LIBS AND ${type} STREQUAL "SHARED")
      install(TARGETS ${libname}
        ARCHIVE DESTINATION ${COMPILER_RT_LIBRARY_INSTALL_DIR}
                ${COMPONENT_OPTION}
        LIBRARY DESTINATION ${COMPILER_RT_LIBRARY_INSTALL_DIR}
                ${COMPONENT_OPTION}
        RUNTIME DESTINATION ${COMPILER_RT_LIBRARY_INSTALL_DIR}
                ${COMPONENT_OPTION})

      # We only want to generate per-library install targets if you aren't using
      # an IDE because the extra targets get cluttered in IDEs.
      if(NOT CMAKE_CONFIGURATION_TYPES)
        add_custom_target(install-${libname}
                          DEPENDS ${libname}
                          COMMAND "${CMAKE_COMMAND}"
                                  -DCMAKE_INSTALL_COMPONENT=${libname}
                                  -P "${CMAKE_BINARY_DIR}/cmake_install.cmake")
        # If you have a parent target specified, we bind the new install target
        # to the parent install target.
        if(LIB_PARENT_TARGET)
          add_dependencies(install-${LIB_PARENT_TARGET} install-${libname})
        endif(LIB_PARENT_TARGET)
      endif(NOT CMAKE_CONFIGURATION_TYPES)

    endforeach(libname)
    if(LIB_PARENT_TARGET)
      add_dependencies(${LIB_PARENT_TARGET} ${libnames})
    endif(LIB_PARENT_TARGET)
  endif(libnames)
endmacro(add_compiler_rt_runtime)

append_list_if(COMPILER_RT_DEBUG -DSANITIZER_DEBUG=1 COMPILER_RT_UNITTEST_CFLAGS)

macro(add_compiler_rt_resource_file target_name file_name component)
  set(src_file "${CMAKE_CURRENT_SOURCE_DIR}/${file_name}")
  set(dst_file "${COMPILER_RT_OUTPUT_DIR}/${file_name}")
  add_custom_command(OUTPUT ${dst_file}
    DEPENDS ${src_file}
    COMMAND ${CMAKE_COMMAND} -E copy_if_different ${src_file} ${dst_file}
    COMMENT "Copying ${file_name}...")
  add_custom_target(${target_name} DEPENDS ${dst_file})
  # Install in Clang resource directory.
  install(FILES ${file_name}
    DESTINATION ${COMPILER_RT_INSTALL_PATH}
    COMPONENT ${component})
  add_dependencies(${component} ${target_name})

endmacro(add_compiler_rt_resource_file)

macro(add_compiler_rt_script name)
  set(dst ${COMPILER_RT_EXEC_OUTPUT_DIR}/${name})
  set(src ${CMAKE_CURRENT_SOURCE_DIR}/${name})
  add_custom_command(OUTPUT ${dst}
    DEPENDS ${src}
    COMMAND ${CMAKE_COMMAND} -E copy_if_different ${src} ${dst}
    COMMENT "Copying ${name}...")
  add_custom_target(${name} DEPENDS ${dst})
  install(FILES ${dst}
    PERMISSIONS OWNER_READ OWNER_WRITE OWNER_EXECUTE GROUP_READ GROUP_EXECUTE WORLD_READ WORLD_EXECUTE
    DESTINATION ${COMPILER_RT_INSTALL_PATH}/bin)
endmacro(add_compiler_rt_script src name)
