include(CMakePushCheckState)
include(CheckSymbolExists)

# Because compiler-rt spends a lot of time setting up custom compile flags,
# define a handy helper macro for it. The compile flags setting in CMake
# has serious issues that make its syntax challenging at best.
macro(set_target_compile_flags target)
  set(argstring "")
  foreach(arg ${ARGN})
    set(argstring "${argstring} ${arg}")
  endforeach(arg)
  set_target_properties(${target} PROPERTIES COMPILE_FLAGS "${argstring}")
endmacro(set_target_compile_flags)

macro(set_target_link_flags target)
  set(argstring "")
  foreach(arg ${ARGN})
    set(argstring "${argstring} ${arg}")
  endforeach(arg)
  set_target_properties(${target} PROPERTIES LINK_FLAGS "${argstring}")
endmacro(set_target_link_flags)

# Appends value to all lists in ARGN, if the condition is true.
macro(append_list_if condition value)
  if(${condition})
    foreach(list ${ARGN})
      list(APPEND ${list} ${value})
    endforeach(list)
  endif(${condition})
endmacro(append_list_if )

macro(list_find user_list key index)
  set(ind 0)
  set(is_found 0)
  set(${index} "-1")

  foreach(val ${${user_list}})
    if ("${val}" STREQUAL "${key}" AND NOT ${is_found} EQUAL 1)
      set(${index} ${ind})
      set(is_found 1)
    endif("${val}" STREQUAL "${key}" AND NOT ${is_found} EQUAL 1)

    math(EXPR ind "${ind} + 1")
  endforeach(val)
endmacro(list_find)

# Appends value to all strings in ARGN, if the condition is true.
macro(append_string_if condition value)
  if(${condition})
    foreach(str ${ARGN})
      set(${str} "${${str}} ${value}")
    endforeach(str)
  endif(${condition})
endmacro(append_string_if)

macro(append_rtti_flag polarity list)
  if(polarity)
    append_list_if(COMPILER_RT_HAS_FRTTI_FLAG -frtti ${list})
    append_list_if(COMPILER_RT_HAS_GR_FLAG /GR ${list})
  else(polarity)
    append_list_if(COMPILER_RT_HAS_FNO_RTTI_FLAG -fno-rtti ${list})
    append_list_if(COMPILER_RT_HAS_GR_FLAG /GR- ${list})
  endif(polarity)
endmacro(append_rtti_flag)

macro(append_have_file_definition filename varname list)
  check_include_file("${filename}" "${varname}")
  if (NOT ${varname})
    set("${varname}" 0)
  endif(NOT ${varname})
  list(APPEND ${list} "${varname}=${${varname}}")
endmacro(append_have_file_definition)

# Takes ${ARGN} and puts only supported architectures in @out_var list.
macro(filter_available_targets out_var)
  set(archs ${${out_var}})
  foreach(arch ${ARGN})
    list_find(COMPILER_RT_SUPPORTED_ARCH ${arch} ARCH_INDEX)
    if(NOT ARCH_INDEX EQUAL -1 AND CAN_TARGET_${arch})
      list(APPEND archs ${arch})
    endif(NOT ARCH_INDEX EQUAL -1 AND CAN_TARGET_${arch})
  endforeach(arch)
  set(${out_var} ${archs})
endmacro(filter_available_targets)

macro(check_compile_definition def argstring out_var)
  if("${def}" STREQUAL "")
    set(${out_var} TRUE)
  else("${def}" STREQUAL "")
    cmake_push_check_state()
    set(CMAKE_REQUIRED_FLAGS "${CMAKE_REQUIRED_FLAGS} ${argstring}")
    check_symbol_exists(${def} "" ${out_var})
    cmake_pop_check_state()
  endif("${def}" STREQUAL "")
endmacro(check_compile_definition)

# test_target_arch(<arch> <def> <target flags...>)
# Checks if architecture is supported: runs host compiler with provided
# flags to verify that:
#   1) <def> is defined (if non-empty)
#   2) simple file can be successfully built.
# If successful, saves target flags for this architecture.
macro(test_target_arch arch def)
  set(TARGET_${arch}_CFLAGS ${ARGN})
  set(TARGET_${arch}_LINKFLAGS ${ARGN})
  set(argstring "")
  foreach(arg ${ARGN})
    set(argstring "${argstring} ${arg}")
  endforeach(arg)
  check_compile_definition("${def}" "${argstring}" HAS_${arch}_DEF)
  if(NOT HAS_${arch}_DEF)
    set(CAN_TARGET_${arch} FALSE)
  elseif(TEST_COMPILE_ONLY)
    try_compile_only(CAN_TARGET_${arch} ${TARGET_${arch}_CFLAGS})
  else(NOT HAS_${arch}_DEF)
    set(argstring "${CMAKE_EXE_LINKER_FLAGS} ${argstring}")
    try_compile(CAN_TARGET_${arch} ${CMAKE_BINARY_DIR} ${SIMPLE_SOURCE}
                COMPILE_DEFINITIONS "${TARGET_${arch}_CFLAGS}"
                OUTPUT_VARIABLE TARGET_${arch}_OUTPUT
                CMAKE_FLAGS "-DCMAKE_EXE_LINKER_FLAGS:STRING=${argstring}")
  endif(NOT HAS_${arch}_DEF)
  if(${CAN_TARGET_${arch}})
    list(APPEND COMPILER_RT_SUPPORTED_ARCH ${arch})
  endif(${CAN_TARGET_${arch}})
endmacro(test_target_arch)

macro(detect_target_arch)
  check_symbol_exists(__arm__ "" __ARM)
  check_symbol_exists(__aarch64__ "" __AARCH64)
  check_symbol_exists(__x86_64__ "" __X86_64)
  check_symbol_exists(__i686__ "" __I686)
  check_symbol_exists(__i386__ "" __I386)
  check_symbol_exists(__mips__ "" __MIPS)
  check_symbol_exists(__mips64__ "" __MIPS64)
  check_symbol_exists(__s390x__ "" __S390X)
  check_symbol_exists(__wasm32__ "" __WEBASSEMBLY32)
  check_symbol_exists(__wasm64__ "" __WEBASSEMBLY64)
  check_symbol_exists(__e2k__ "" __E2K) # MCST
  check_symbol_exists(__sparc__ "" __SPARC)
  if(__ARM)
    add_default_target_arch(arm)
  elseif(__AARCH64)
    add_default_target_arch(aarch64)
  elseif(__X86_64)
    add_default_target_arch(x86_64)
  elseif(__I686)
    add_default_target_arch(i686)
  elseif(__I386)
    add_default_target_arch(i386)
  elseif(__MIPS64) # must be checked before __MIPS
    add_default_target_arch(mips64)
  elseif(__MIPS)
    add_default_target_arch(mips)
  elseif(__S390X)
    add_default_target_arch(s390x)
  elseif(__WEBASSEMBLY32)
    add_default_target_arch(wasm32)
  elseif(__WEBASSEMBLY64)
    add_default_target_arch(wasm64)
  elseif(__E2K)
    add_default_target_arch(e2k) # MCST
  elseif(__SPARC)
    add_default_target_arch(sparc)
  endif(__ARM)
endmacro(detect_target_arch)
