# The CompilerRT build system requires CMake version 2.8.8 or higher in order
# to use its support for building convenience "libraries" as a collection of
# .o files. This is particularly useful in producing larger, more complex
# runtime libraries.

include(CheckIncludeFile)
check_include_file(unwind.h HAVE_UNWIND_H)

# Top level target used to build all compiler-rt libraries.
#add_custom_target(compiler-rt ALL)

# Take output dir and install path from the user.
set(COMPILER_RT_OUTPUT_DIR ${CMAKE_CURRENT_BINARY_DIR} CACHE PATH
  "Path where built compiler-rt libraries should be stored.")
set(COMPILER_RT_EXEC_OUTPUT_DIR ${CMAKE_CURRENT_BINARY_DIR}/bin CACHE PATH
  "Path where built compiler-rt executables should be stored.")
set(COMPILER_RT_INSTALL_PATH ${CMAKE_INSTALL_PREFIX} CACHE PATH
  "Path where built compiler-rt libraries should be installed.")
option(COMPILER_RT_INCLUDE_TESTS "Generate and build compiler-rt unit tests." OFF)
option(COMPILER_RT_ENABLE_WERROR "Fail and stop if warning is triggered" OFF)
# Use a host compiler to compile/link tests.
set(COMPILER_RT_TEST_COMPILER ${CMAKE_C_COMPILER} CACHE PATH "Compiler to use for testing")
set(COMPILER_RT_TEST_CXX_COMPILER ${CMAKE_CXX_COMPILER} CACHE PATH "C++ Compiler to use for testing")

string(TOLOWER ${CMAKE_SYSTEM_NAME} COMPILER_RT_OS_DIR)
set(COMPILER_RT_LIBRARY_OUTPUT_DIR
  ${COMPILER_RT_OUTPUT_DIR}/lib/${COMPILER_RT_OS_DIR})
if (MCST)
  # MCST: compiler-rt is always built via extern script that sets common
  # path for all libraries and them headers that have common path given
  # to all binaries and headers of libraries.
  if (NOT MCST_SANITIZE_LIBRARY_INSTALL_DIR OR 
      NOT MCST_SANITIZE_HEADER_INSTALL_DIR)
      message(FATAL_ERROR "installation directories for libraries and headers must be set")
  endif(NOT MCST_SANITIZE_LIBRARY_INSTALL_DIR OR 
        NOT MCST_SANITIZE_HEADER_INSTALL_DIR)
 # if (NOT IS_ABSOLUTE ${MCST_SANITIZE_LIBRARY_INSTALL_DIR} OR
 #     NOT IS_ABSOLUTE ${MCST_SANITIZE_HEADER_INSTALL_DIR})
 #     message(FATAL_ERROR "installation paths must be absolute")
 # endif(NOT IS_ABSOLUTE ${MCST_SANITIZE_LIBRARY_INSTALL_DIR} OR
 #       NOT IS_ABSOLUTE ${MCST_SANITIZE_HEADER_INSTALL_DIR})
  set(COMPILER_RT_LIBRARY_INSTALL_DIR ${MCST_SANITIZE_LIBRARY_INSTALL_DIR})
  set(COMPILER_RT_HEADER_INSTALL_DIR ${MCST_SANITIZE_HEADER_INSTALL_DIR})
else(MCST)
  set(COMPILER_RT_LIBRARY_INSTALL_DIR
    ${COMPILER_RT_INSTALL_PATH}/lib/${COMPILER_RT_OS_DIR})
endif(MCST)

macro(test_targets)
  # Generate the COMPILER_RT_SUPPORTED_ARCH list.
  if(ANDROID)
    # Examine compiler output to determine target architecture.
    detect_target_arch()
    set(COMPILER_RT_OS_SUFFIX "-android")
  elseif(MCST)
    # MCST: we don't use llvm-config so TARGET_TRIPLE is not set 
    # and we must test target architecture explicitly
    # TODO: in future we need to add e2k64 and sparc64
    #test_target_arch(e2k __e2k__)
    #test_target_arch(sparc __sparc__)
    detect_target_arch()
  elseif(NOT APPLE) # Supported archs for Apple platforms are generated later
    if("${COMPILER_RT_DEFAULT_TARGET_ARCH}" MATCHES "i[2-6]86|x86|amd64")
      if(NOT MSVC)
        test_target_arch(x86_64 "" "-m64")
        # FIXME: We build runtimes for both i686 and i386, as "clang -m32" may
        # target different variant than "$CMAKE_C_COMPILER -m32". This part should
        # be gone after we resolve PR14109.
        test_target_arch(i686 __i686__ "-m32")
        test_target_arch(i386 __i386__ "-m32")
      else(NOT MSVC)
        if (CMAKE_SIZEOF_VOID_P EQUAL 4)
          test_target_arch(i386 "" "")
        else(CMAKE_SIZEOF_VOID_P EQUAL 4)
          test_target_arch(x86_64 "" "")
        endif(CMAKE_SIZEOF_VOID_P EQUAL 4)
      endif(NOT MSVC)
    elseif("${COMPILER_RT_DEFAULT_TARGET_ARCH}" MATCHES "powerpc")
      TEST_BIG_ENDIAN(HOST_IS_BIG_ENDIAN)
      if(HOST_IS_BIG_ENDIAN)
        test_target_arch(powerpc64 "" "-m64")
      else(HOST_IS_BIG_ENDIAN)
        test_target_arch(powerpc64le "" "-m64")
      endif(HOST_IS_BIG_ENDIAN)
    elseif("${COMPILER_RT_DEFAULT_TARGET_ARCH}" MATCHES "s390x")
      test_target_arch(s390x "" "")
    elseif("${COMPILER_RT_DEFAULT_TARGET_ARCH}" MATCHES "mipsel|mips64el")
      # Gcc doesn't accept -m32/-m64 so we do the next best thing and use
      # -mips32r2/-mips64r2. We don't use -mips1/-mips3 because we want to match
      # clang's default CPU's. In the 64-bit case, we must also specify the ABI
      # since the default ABI differs between gcc and clang.
      # FIXME: Ideally, we would build the N32 library too.
      test_target_arch(mipsel "" "-mips32r2" "--target=mipsel-linux-gnu")
      test_target_arch(mips64el "" "-mips64r2" "--target=mips64el-linux-gnu" "-mabi=64")
    elseif("${COMPILER_RT_DEFAULT_TARGET_ARCH}" MATCHES "mips")
      test_target_arch(mips "" "-mips32r2" "--target=mips-linux-gnu")
      test_target_arch(mips64 "" "-mips64r2" "--target=mips64-linux-gnu" "-mabi=64")
    elseif("${COMPILER_RT_DEFAULT_TARGET_ARCH}" MATCHES "arm")
      test_target_arch(arm "" "-march=armv7-a" "-mfloat-abi=soft")
      test_target_arch(armhf "" "-march=armv7-a" "-mfloat-abi=hard")
    elseif("${COMPILER_RT_DEFAULT_TARGET_ARCH}" MATCHES "aarch32")
      test_target_arch(aarch32 "" "-march=armv8-a")
    elseif("${COMPILER_RT_DEFAULT_TARGET_ARCH}" MATCHES "aarch64")
      test_target_arch(aarch64 "" "-march=armv8-a")
    elseif("${COMPILER_RT_DEFAULT_TARGET_ARCH}" MATCHES "wasm32")
      test_target_arch(wasm32 "" "--target=wasm32-unknown-unknown")
    elseif("${COMPILER_RT_DEFAULT_TARGET_ARCH}" MATCHES "wasm64")
      test_target_arch(wasm64 "" "--target=wasm64-unknown-unknown")
    endif("${COMPILER_RT_DEFAULT_TARGET_ARCH}" MATCHES "i[2-6]86|x86|amd64")
    set(COMPILER_RT_OS_SUFFIX "")
  endif(ANDROID)
endmacro(test_targets)
